/* Copyright (c) 2004 The Nutch Organization.  All rights reserved.   */
/* Use subject to the conditions in http://www.nutch.org/LICENSE.txt. */

package net.nutch.protocol.ftp;

import javax.activation.MimetypesFileTypeMap;
// 20040528, xing, disabled for now
//import xing.net.nutch.util.magicfile.*;

import org.apache.commons.net.ftp.FTPFileEntryParser;

import net.nutch.net.protocols.HttpDateFormat;

import net.nutch.util.LogFormatter;
import net.nutch.util.NutchConf;

import net.nutch.protocol.Content;
import net.nutch.protocol.Protocol;

import java.util.logging.Level;
import java.util.logging.Logger;

import java.net.URL;

import java.io.InputStream;
// 20040528, xing, disabled for now
//import java.io.Reader;
import java.io.IOException;

/************************************
 * Ftp.java deals with ftp: scheme.
 *
 * Configurable parameters are defined under "FTP properties" section
 * in ./conf/nutch-default.xml or similar.
 *
 * @author John Xing
 ***********************************/
public class Ftp implements Protocol {

  public static final Logger LOG =
    LogFormatter.getLogger("net.nutch.protocol.ftp.Ftp");

  static final int BUFFER_SIZE = 16384; // 16*1024 = 16384

  static final int MAX_REDIRECTS = 5;

  static int timeout = NutchConf.getInt("ftp.timeout", 10000);

  static int maxContentLength = NutchConf.getInt("ftp.content.limit",64*1024);

  String userName = NutchConf.get("ftp.username", "anonymous");
  String passWord = NutchConf.get("ftp.password", "anonymous@example.com");

  // typical/default server timeout is 120*1000 millisec.
  // better be conservative here
  int serverTimeout = NutchConf.getInt("ftp.server.timeout", 60*1000);

  // when to have client start anew
  long renewalTime = -1;

  boolean keepConnection = NutchConf.getBoolean("ftp.keep.connection", false);

  boolean followTalk = NutchConf.getBoolean("ftp.follow.talk", false);

  // ftp client
  Client client = null;
  // ftp dir list entry parser
  FTPFileEntryParser parser = null;

  // 20040412, xing
  // the following three: HttpDateFormat, MimetypesFileTypeMap, MagicFile
  // are placed in each thread before we check out if they're thread-safe.

  // http date format
  HttpDateFormat httpDateFormat = null;

  // file name extension to mime-type map
  static MimetypesFileTypeMap TYPE_MAP = null;

  static {
    try {
      // read mime types from config file
      InputStream is =
        NutchConf.getConfResourceAsInputStream
        (NutchConf.get("mime.types.file"));
      if (is == null) {
        LOG.warning
          ("no mime.types.file: won't use url extension for content-type.");
        TYPE_MAP = null;
      } else {
        TYPE_MAP = new MimetypesFileTypeMap(is);
      }
      
      if (is != null)
        is.close();
    } catch (IOException e) {
      LOG.log(Level.SEVERE, "Unexpected error", e);
    }
  }

// 20040528, xing, disabled for now
//  // file magic for determining content type
//  static MagicFile MAGIC = null;
//
//  static {
//    try {
//      // read file magic from config file
//      Reader reader =
//        NutchConf.getConfResourceAsReader
//          (NutchConf.get("mime.magic.file"));
//      if (reader == null) {
//        LOG.warning
//          ("no mime.magic.file: won't use file magic for content-type.");
//        MAGIC = null;
//      } else {
//        MAGIC = MagicFile.getInstance(reader);
//      }
//
//      if (reader != null)
//        reader.close();
//    } catch (IOException e) {
//      LOG.log(Level.SEVERE, "Unexpected error", e);
//    }
//  }

  // constructor
  public Ftp() {
    this.httpDateFormat = new HttpDateFormat();
  }

  /** Set the timeout. */
  public void setTimeout(int timeout) {
    this.timeout = timeout;
  }

  /** Set the point at which content is truncated. */
  public void setMaxContentLength(int length) {
    this.maxContentLength = length;
  }

  /** Set followTalk */
  public void setFollowTalk(boolean followTalk) {
    this.followTalk = followTalk;
  }

  /** Set keepConnection */
  public void setKeepConnection(boolean keepConnection) {
    this.keepConnection = keepConnection;
  }

  public Content getContent(String urlString) throws FtpException {
    try {
      URL url = new URL(urlString);
  
      int redirects = 0;
  
      while (true) {
        FtpResponse response;
        response = new FtpResponse(urlString, url, this);   // make a request
  
        int code = response.getCode();
  
        if (code == 200) {                          // got a good response
          return response.toContent();              // return it
  
        } else if (code >= 300 && code < 400) {     // handle redirect
          if (redirects == MAX_REDIRECTS)
            throw new FtpException("Too many redirects: " + url);
          url = new URL(response.getHeader("Location"));
          redirects++;                
          if (LOG.isLoggable(Level.FINE))
            LOG.fine("redirect to " + url); 
  
        } else {                                    // convert to exception
          throw new FtpError(code);
        }
      } 
    } catch (IOException e) {
      throw new FtpException(e);
    }
  }

  protected void finalize () {
    try {
      if (this.client != null && this.client.isConnected()) {
        this.client.logout();
        this.client.disconnect();
      }
    } catch (IOException e) {
      // do nothing
    }
  }

  /** For debugging. */
  public static void main(String[] args) throws Exception {
    int timeout = Integer.MIN_VALUE;
    int maxContentLength = Integer.MIN_VALUE;
    String logLevel = "info";
    boolean followTalk = false;
    boolean keepConnection = false;
    boolean dumpContent = false;
    String urlString = null;

    String usage = "Usage: Ftp [-logLevel level] [-followTalk] [-keepConnection] [-timeout N] [-maxContentLength L] [-dumpContent] url";

    if (args.length == 0) {
      System.err.println(usage);
      System.exit(-1);
    }
      
    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-logLevel")) {
        logLevel = args[++i];
      } else if (args[i].equals("-followTalk")) {
        followTalk = true;
      } else if (args[i].equals("-keepConnection")) {
        keepConnection = true;
      } else if (args[i].equals("-timeout")) {
        timeout = Integer.parseInt(args[++i]) * 1000;
      } else if (args[i].equals("-maxContentLength")) {
        maxContentLength = Integer.parseInt(args[++i]);
      } else if (args[i].equals("-dumpContent")) {
        dumpContent = true;
      } else if (i != args.length-1) {
        System.err.println(usage);
        System.exit(-1);
      } else {
        urlString = args[i];
      }
    }

    Ftp ftp = new Ftp();

    ftp.setFollowTalk(followTalk);
    ftp.setKeepConnection(keepConnection);

    if (timeout != Integer.MIN_VALUE) // set timeout
      ftp.setTimeout(timeout);

    if (maxContentLength != Integer.MIN_VALUE) // set maxContentLength
      ftp.setMaxContentLength(maxContentLength);

    // set log level
    LOG.setLevel(Level.parse((new String(logLevel)).toUpperCase()));

    Content content = ftp.getContent(urlString);

    System.err.println("Content-Type: " + content.getContentType());
    System.err.println("Content-Length: " + content.get("Content-Length"));
    System.err.println("Last-Modified: " + content.get("Last-Modified"));
    if (dumpContent) {
      System.out.print(new String(content.getContent()));
    }

    ftp = null;
  }

}
